@@ -76,67 +76,21 @@ display_control_sequence(const char *name, GValueArray *params)
 /* A couple are duplicated from vte.c, to keep them static... */
 
 /* Find the character an the given position in the backscroll buffer. */
-static struct vte_charcell *
-vte_terminal_find_charcell(VteTerminal *terminal, glong col, glong row)
+static VteCell *
+vte_terminal_find_charcell (VteTerminal *terminal, glong col, glong row)
 {
 	VteRowData *rowdata;
-	struct vte_charcell *ret = NULL;
+	VteCell *ret = NULL;
 	VteScreen *screen;
 	g_assert(VTE_IS_TERMINAL(terminal));
 	screen = terminal->pvt->screen;
-	if (_vte_ring_contains(screen->row_data, row)) {
-		rowdata = _vte_ring_index(screen->row_data, row);
-		if ((glong) rowdata->cells->len > col) {
-			ret = &g_array_index(rowdata->cells,
-					     struct vte_charcell,
-					     col);
-		}
+	if (_vte_ring_contains (screen->row_data, row)) {
+		rowdata = _vte_ring_index_writable (screen->row_data, row);
+		ret = _vte_row_data_get_writable (rowdata, col);
 	}
 	return ret;
 }
 
-/* Append a single item to a GArray a given number of times. Centralizing all
- * of the places we do this may let me do something more clever later.
- * Dupped from vte.c. */
-static void
-vte_g_array_fill(GArray *array, gpointer item, guint final_size)
-{
-	g_assert(array != NULL);
-	if (array->len >= final_size) {
-		return;
-	}
-	g_assert(item != NULL);
-
-	final_size -= array->len;
-	do {
-		g_array_append_vals(array, item, 1);
-	} while (--final_size);
-}
-
-/* Insert a blank line at an arbitrary position. */
-static void
-vte_insert_line_internal(VteTerminal *terminal, glong position)
-{
-	/* Pad out the line data to the insertion point. */
-	while (_vte_ring_next(terminal->pvt->screen->row_data) < position) {
-		_vte_ring_append(terminal->pvt->screen->row_data);
-	}
-	/* If we haven't inserted a line yet, insert a new one. */
-	if (_vte_ring_next(terminal->pvt->screen->row_data) >= position) {
-		_vte_ring_insert(terminal->pvt->screen->row_data, position);
-	} else {
-		_vte_ring_append(terminal->pvt->screen->row_data);
-	}
-}
-
-/* Remove a line at an arbitrary position. */
-static void
-vte_remove_line_internal(VteTerminal *terminal, glong position)
-{
-	if (_vte_ring_next(terminal->pvt->screen->row_data) > position)
-		_vte_ring_remove(terminal->pvt->screen->row_data, position);
-}
-
 /* Check how long a string of unichars is.  Slow version. */
 static gssize
 vte_unichar_strlen(gunichar *c)
@@ -323,7 +277,7 @@ _vte_terminal_clear_screen (VteTerminal *terminal)
 	initial = _vte_ring_next(screen->row_data);
 	/* Add a new screen's worth of rows. */
 	for (i = 0; i < terminal->row_count; i++)
-		_vte_ring_append(screen->row_data);
+		_vte_terminal_ring_append (terminal, TRUE);
 	/* Move the cursor and insertion delta to the first line in the
 	 * newly-cleared area and scroll if need be. */
 	screen->insert_delta = initial;
@@ -348,18 +302,13 @@ _vte_terminal_clear_current_line (VteTerminal *terminal)
 	 * which corresponds to the cursor. */
 	if (_vte_ring_next(screen->row_data) > screen->cursor_current.row) {
 		/* Get the data for the row which the cursor points to. */
-		rowdata = _vte_ring_index(screen->row_data, screen->cursor_current.row);
+		rowdata = _vte_ring_index_writable (screen->row_data, screen->cursor_current.row);
 		g_assert(rowdata != NULL);
 		/* Remove it. */
-		if (rowdata->cells->len > 0) {
-			g_array_set_size(rowdata->cells, 0);
-		}
-		/* Add enough cells to the end of the line to fill out the
-		 * row. */
-		vte_g_array_fill(rowdata->cells,
-				 &screen->fill_defaults,
-				 terminal->column_count);
-		rowdata->soft_wrapped = 0;
+		_vte_row_data_shrink (rowdata, 0);
+		/* Add enough cells to the end of the line to fill out the row. */
+		_vte_row_data_fill (rowdata, &screen->fill_defaults, terminal->column_count);
+		rowdata->attr.soft_wrapped = 0;
 		/* Repaint this row. */
 		_vte_invalidate_cells(terminal,
 				      0, terminal->column_count,
@@ -382,20 +331,14 @@ _vte_terminal_clear_above_current (VteTerminal *terminal)
 	 * which corresponds to the cursor. */
 	for (i = screen->insert_delta; i < screen->cursor_current.row; i++) {
 		if (_vte_ring_next(screen->row_data) > i) {
-			guint len;
 			/* Get the data for the row we're erasing. */
-			rowdata = _vte_ring_index(screen->row_data, i);
+			rowdata = _vte_ring_index_writable (screen->row_data, i);
 			g_assert(rowdata != NULL);
 			/* Remove it. */
-			len = rowdata->cells->len;
-			if (len > 0) {
-				g_array_set_size(rowdata->cells, 0);
-			}
+			_vte_row_data_shrink (rowdata, 0);
 			/* Add new cells until we fill the row. */
-			vte_g_array_fill(rowdata->cells,
-					 &screen->fill_defaults,
-					 terminal->column_count);
-			rowdata->soft_wrapped = 0;
+			_vte_row_data_fill (rowdata, &screen->fill_defaults, terminal->column_count);
+			rowdata->attr.soft_wrapped = 0;
 			/* Repaint the row. */
 			_vte_invalidate_cells(terminal,
 					0, terminal->column_count, i, 1);
@@ -423,17 +366,17 @@ _vte_terminal_scroll_text (VteTerminal *terminal, int scroll_amount)
 	}
 
 	while (_vte_ring_next(screen->row_data) <= end)
-		_vte_ring_append(terminal->pvt->screen->row_data);
+		_vte_terminal_ring_append (terminal, FALSE);
 
 	if (scroll_amount > 0) {
 		for (i = 0; i < scroll_amount; i++) {
-			vte_remove_line_internal(terminal, end);
-			vte_insert_line_internal(terminal, start);
+			_vte_terminal_ring_remove (terminal, end);
+			_vte_terminal_ring_insert (terminal, start, TRUE);
 		}
 	} else {
 		for (i = 0; i < -scroll_amount; i++) {
-			vte_remove_line_internal(terminal, start);
-			vte_insert_line_internal(terminal, end);
+			_vte_terminal_ring_remove (terminal, start);
+			_vte_terminal_ring_insert (terminal, end, TRUE);
 		}
 	}
 
@@ -954,7 +897,6 @@ static void
 vte_sequence_handler_al (VteTerminal *terminal, GValueArray *params)
 {
 	VteScreen *screen;
-	VteRowData *rowdata;
 	long start, end, param, i;
 	GValue *value;
 
@@ -980,14 +922,8 @@ vte_sequence_handler_al (VteTerminal *terminal, GValueArray *params)
 	for (i = 0; i < param; i++) {
 		/* Clear a line off the end of the region and add one to the
 		 * top of the region. */
-		vte_remove_line_internal(terminal, end);
-		vte_insert_line_internal(terminal, start);
-		/* Get the data for the new row. */
-		rowdata = _vte_ring_index(screen->row_data, start);
-		g_assert(rowdata != NULL);
-		/* Add enough cells to it so that it has the default columns. */
-		vte_g_array_fill(rowdata->cells, &screen->fill_defaults,
-				 terminal->column_count);
+		_vte_terminal_ring_remove (terminal, end);
+		_vte_terminal_ring_insert (terminal, start, TRUE);
 		/* Adjust the scrollbars if necessary. */
 		_vte_terminal_adjust_adjustments(terminal);
 	}
@@ -1059,7 +995,7 @@ vte_sequence_handler_cb (VteTerminal *terminal, GValueArray *params)
 	VteRowData *rowdata;
 	long i;
 	VteScreen *screen;
-	struct vte_charcell *pcell;
+	VteCell *pcell;
 	screen = terminal->pvt->screen;
 
 	/* Get the data for the row which the cursor points to. */
@@ -1068,16 +1004,13 @@ vte_sequence_handler_cb (VteTerminal *terminal, GValueArray *params)
 	 * attributes.  If there is no such character cell, we need
 	 * to add one. */
 	for (i = 0; i <= screen->cursor_current.col; i++) {
-		if (i < (glong) rowdata->cells->len) {
+		if (i < (glong) _vte_row_data_length (rowdata)) {
 			/* Muck with the cell in this location. */
-			pcell = &g_array_index(rowdata->cells,
-					       struct vte_charcell,
-					       i);
+			pcell = _vte_row_data_get_writable (rowdata, i);
 			*pcell = screen->color_defaults;
 		} else {
 			/* Add new cells until we have one here. */
-			g_array_append_val(rowdata->cells,
-					   screen->color_defaults);
+			_vte_row_data_append (rowdata, &screen->color_defaults);
 		}
 	}
 	/* Repaint this row. */
@@ -1103,41 +1036,35 @@ vte_sequence_handler_cd (VteTerminal *terminal, GValueArray *params)
 	i = screen->cursor_current.row;
 	if (i < _vte_ring_next(screen->row_data)) {
 		/* Get the data for the row we're clipping. */
-		rowdata = _vte_ring_index(screen->row_data, i);
+		rowdata = _vte_ring_index_writable (screen->row_data, i);
 		/* Clear everything to the right of the cursor. */
-		if ((rowdata != NULL) &&
-		    ((glong) rowdata->cells->len > screen->cursor_current.col)) {
-			g_array_set_size(rowdata->cells,
-					 screen->cursor_current.col);
-		}
+		if (rowdata)
+			_vte_row_data_shrink (rowdata, screen->cursor_current.col);
 	}
 	/* Now for the rest of the lines. */
 	for (i = screen->cursor_current.row + 1;
 	     i < _vte_ring_next(screen->row_data);
 	     i++) {
 		/* Get the data for the row we're removing. */
-		rowdata = _vte_ring_index(screen->row_data, i);
+		rowdata = _vte_ring_index_writable (screen->row_data, i);
 		/* Remove it. */
-		if ((rowdata != NULL) && (rowdata->cells->len > 0)) {
-			g_array_set_size(rowdata->cells, 0);
-		}
+		if (rowdata)
+			_vte_row_data_shrink (rowdata, 0);
 	}
 	/* Now fill the cleared areas. */
 	for (i = screen->cursor_current.row;
 	     i < screen->insert_delta + terminal->row_count;
 	     i++) {
 		/* Retrieve the row's data, creating it if necessary. */
-		if (_vte_ring_contains(screen->row_data, i)) {
-			rowdata = _vte_ring_index(screen->row_data, i);
+		if (_vte_ring_contains (screen->row_data, i)) {
+			rowdata = _vte_ring_index_writable (screen->row_data, i);
 			g_assert(rowdata != NULL);
 		} else {
-			rowdata = _vte_ring_append(screen->row_data);
+			rowdata = _vte_terminal_ring_append (terminal, FALSE);
 		}
 		/* Pad out the row. */
-		vte_g_array_fill(rowdata->cells,
-				 &screen->fill_defaults,
-				 terminal->column_count);
-		rowdata->soft_wrapped = 0;
+		_vte_row_data_fill (rowdata, &screen->fill_defaults, terminal->column_count);
+		rowdata->attr.soft_wrapped = 0;
 		/* Repaint this row. */
 		_vte_invalidate_cells(terminal,
 				      0, terminal->column_count,
@@ -1161,18 +1088,16 @@ vte_sequence_handler_ce (VteTerminal *terminal, GValueArray *params)
 	g_assert(rowdata != NULL);
 	/* Remove the data at the end of the array until the current column
 	 * is the end of the array. */
-	if ((glong) rowdata->cells->len > screen->cursor_current.col) {
-		g_array_set_size(rowdata->cells, screen->cursor_current.col);
+	if ((glong) _vte_row_data_length (rowdata) > screen->cursor_current.col) {
+		_vte_row_data_shrink (rowdata, screen->cursor_current.col);
 		/* We've modified the display.  Make a note of it. */
 		terminal->pvt->text_deleted_flag = TRUE;
 	}
 	if (screen->fill_defaults.attr.back != VTE_DEF_BG) {
 		/* Add enough cells to fill out the row. */
-		vte_g_array_fill(rowdata->cells,
-				 &screen->fill_defaults,
-				 terminal->column_count);
+		_vte_row_data_fill (rowdata, &screen->fill_defaults, terminal->column_count);
 	}
-	rowdata->soft_wrapped = 0;
+	rowdata->attr.soft_wrapped = 0;
 	/* Repaint this row. */
 	_vte_invalidate_cells(terminal,
 			      screen->cursor_current.col,
@@ -1422,17 +1347,15 @@ vte_sequence_handler_dc (VteTerminal *terminal, GValueArray *params)
 	if (_vte_ring_next(screen->row_data) > screen->cursor_current.row) {
 		long len;
 		/* Get the data for the row which the cursor points to. */
-		rowdata = _vte_ring_index(screen->row_data, screen->cursor_current.row);
+		rowdata = _vte_ring_index_writable (screen->row_data, screen->cursor_current.row);
 		g_assert(rowdata != NULL);
 		col = screen->cursor_current.col;
-		len = rowdata->cells->len;
+		len = _vte_row_data_length (rowdata);
 		/* Remove the column. */
 		if (col < len) {
-			g_array_remove_index(rowdata->cells, col);
+			_vte_row_data_remove (rowdata, col);
 			if (screen->fill_defaults.attr.back != VTE_DEF_BG) {
-				vte_g_array_fill (rowdata->cells,
-						&screen->fill_defaults,
-						terminal->column_count);
+				_vte_row_data_fill (rowdata, &screen->fill_defaults, terminal->column_count);
 				len = terminal->column_count;
 			}
 			/* Repaint this row. */
@@ -1483,8 +1406,8 @@ vte_sequence_handler_dl (VteTerminal *terminal, GValueArray *params)
 	for (i = 0; i < param; i++) {
 		/* Clear a line off the end of the region and add one to the
 		 * top of the region. */
-		vte_remove_line_internal(terminal, start);
-		vte_insert_line_internal(terminal, end);
+		_vte_terminal_ring_remove (terminal, start);
+		_vte_terminal_ring_insert (terminal, end, TRUE);
 		/* Adjust the scrollbars if necessary. */
 		_vte_terminal_adjust_adjustments(terminal);
 	}
@@ -1546,7 +1469,7 @@ vte_sequence_handler_ec (VteTerminal *terminal, GValueArray *params)
 	VteScreen *screen;
 	VteRowData *rowdata;
 	GValue *value;
-	struct vte_charcell *cell;
+	VteCell *cell;
 	long col, i, count;
 
 	screen = terminal->pvt->screen;
@@ -1569,18 +1492,14 @@ vte_sequence_handler_ec (VteTerminal *terminal, GValueArray *params)
 		for (i = 0; i < count; i++) {
 			col = screen->cursor_current.col + i;
 			if (col >= 0) {
-				if (col < (glong) rowdata->cells->len) {
+				if (col < (glong) _vte_row_data_length (rowdata)) {
 					/* Replace this cell with the current
 					 * defaults. */
-					cell = &g_array_index(rowdata->cells,
-							      struct vte_charcell,
-							      col);
+					cell = _vte_row_data_get_writable (rowdata, col);
 					*cell = screen->color_defaults;
 				} else {
 					/* Add new cells until we have one here. */
-					vte_g_array_fill(rowdata->cells,
-							 &screen->color_defaults,
-							 col);
+					_vte_row_data_fill (rowdata, &screen->color_defaults, col);
 				}
 			}
 		}
@@ -2056,8 +1975,8 @@ vte_sequence_handler_sr (VteTerminal *terminal, GValueArray *params)
 	if (screen->cursor_current.row == start) {
 		/* If we're at the top of the scrolling region, add a
 		 * line at the top to scroll the bottom off. */
-		vte_remove_line_internal(terminal, end);
-		vte_insert_line_internal(terminal, start);
+		_vte_terminal_ring_remove (terminal, end);
+		_vte_terminal_ring_insert (terminal, start, TRUE);
 		/* Update the display. */
 		_vte_terminal_scroll_region(terminal, start, end - start + 1, 1);
 		_vte_invalidate_cells(terminal,
@@ -2135,26 +2054,23 @@ vte_sequence_handler_ta (VteTerminal *terminal, GValueArray *params)
 		 */
 
 		/* Get rid of trailing empty cells: bug 545924 */
-		if ((glong) rowdata->cells->len > col)
+		if ((glong) _vte_row_data_length (rowdata) > col)
 		{
-			struct vte_charcell *cell;
+			const VteCell *cell;
 			guint i;
-			for (i = rowdata->cells->len; (glong) i > col; i--) {
-				cell = &g_array_index(rowdata->cells,
-						      struct vte_charcell, i - 1);
+			for (i = _vte_row_data_length (rowdata); (glong) i > col; i--) {
+				cell = _vte_row_data_get (rowdata, i - 1);
 				if (cell->attr.fragment || cell->c != 0)
 					break;
 			}
-			g_array_set_size(rowdata->cells, i);
+			_vte_row_data_shrink (rowdata, i);
 		}
 
-		if ((glong) rowdata->cells->len <= col)
+		if ((glong) _vte_row_data_length (rowdata) <= col)
 		  {
-		    struct vte_charcell cell;
+		    VteCell cell;
 
-		    vte_g_array_fill (rowdata->cells,
-				      &screen->fill_defaults,
-				      col);
+		    _vte_row_data_fill (rowdata, &screen->fill_defaults, col);
 
 		    cell.attr = screen->fill_defaults.attr;
 		    cell.attr.invisible = 1; /* FIXME: bug 499944 */
@@ -2167,20 +2083,16 @@ vte_sequence_handler_ta (VteTerminal *terminal, GValueArray *params)
 		       goto fallback_tab;
 		      }
 		    cell.c = '\t';
-		    g_array_append_vals(rowdata->cells, &cell, 1);
+		    _vte_row_data_append (rowdata, &cell);
 
-		    cell.attr = screen->fill_defaults.attr;
+		    cell.attr.columns = 1;
 		    cell.attr.fragment = 1;
-		    vte_g_array_fill (rowdata->cells,
-				      &cell,
-				newcol);
+		    _vte_row_data_fill (rowdata, &cell, newcol);
 		  }
 		else
 		  {
 		  fallback_tab:
-		    vte_g_array_fill (rowdata->cells,
-				      &screen->fill_defaults,
-				      newcol);
+		    _vte_row_data_fill (rowdata, &screen->fill_defaults, newcol);
 		  }
 
 		_vte_invalidate_cells (terminal,
@@ -2229,20 +2141,16 @@ vte_sequence_handler_ts (VteTerminal *terminal, GValueArray *params)
 static void
 vte_sequence_handler_uc (VteTerminal *terminal, GValueArray *params)
 {
-	struct vte_charcell *cell;
+	VteCell *cell;
 	int column;
 	VteScreen *screen;
 
 	screen = terminal->pvt->screen;
 	column = screen->cursor_current.col;
-	cell = vte_terminal_find_charcell(terminal,
-					  column,
-					  screen->cursor_current.row);
+	cell = vte_terminal_find_charcell(terminal, column, screen->cursor_current.row);
 	while ((cell != NULL) && (cell->attr.fragment) && (column > 0)) {
 		column--;
-		cell = vte_terminal_find_charcell(terminal,
-						  column,
-						  screen->cursor_current.row);
+		cell = vte_terminal_find_charcell(terminal, column, screen->cursor_current.row);
 	}
 	if (cell != NULL) {
 		/* Set this character to be underlined. */
@@ -2823,7 +2731,6 @@ vte_sequence_handler_insert_blank_characters (VteTerminal *terminal, GValueArray
 static void
 vte_sequence_handler_insert_lines (VteTerminal *terminal, GValueArray *params)
 {
-	VteRowData *rowdata;
 	GValue *value;
 	VteScreen *screen;
 	long param, end, row;
@@ -2849,15 +2756,8 @@ vte_sequence_handler_insert_lines (VteTerminal *terminal, GValueArray *params)
 	for (i = 0; i < param; i++) {
 		/* Clear a line off the end of the region and add one to the
 		 * top of the region. */
-		vte_remove_line_internal(terminal, end);
-		vte_insert_line_internal(terminal, row);
-		/* Get the data for the new row. */
-		rowdata = _vte_ring_index(screen->row_data, row);
-		g_assert(rowdata != NULL);
-		/* Add enough cells to it so that it has the default colors. */
-		vte_g_array_fill(rowdata->cells,
-				 &screen->fill_defaults,
-				 terminal->column_count);
+		_vte_terminal_ring_remove (terminal, end);
+		_vte_terminal_ring_insert (terminal, row, TRUE);
 	}
 	/* Update the display. */
 	_vte_terminal_scroll_region(terminal, row, end - row + 1, param);
@@ -2872,7 +2772,6 @@ static void
 vte_sequence_handler_delete_lines (VteTerminal *terminal, GValueArray *params)
 {
 	GValue *value;
-	VteRowData *rowdata;
 	VteScreen *screen;
 	long param, end, row;
 	int i;
@@ -2898,15 +2797,8 @@ vte_sequence_handler_delete_lines (VteTerminal *terminal, GValueArray *params)
 	for (i = 0; i < param; i++) {
 		/* Insert a line at the end of the region and remove one from
 		 * the top of the region. */
-		vte_remove_line_internal(terminal, row);
-		vte_insert_line_internal(terminal, end);
-		/* Get the data for the new row. */
-		rowdata = _vte_ring_index(screen->row_data, end);
-		g_assert(rowdata != NULL);
-		/* Add enough cells to it so that it has the default colors. */
-		vte_g_array_fill(rowdata->cells,
-				 &screen->fill_defaults,
-				 terminal->column_count);
+		_vte_terminal_ring_remove (terminal, row);
+		_vte_terminal_ring_insert (terminal, end, TRUE);
 	}
 	/* Update the display. */
 	_vte_terminal_scroll_region(terminal, row, end - row + 1, -param);
@@ -3063,7 +2955,7 @@ vte_sequence_handler_screen_alignment_test (VteTerminal *terminal, GValueArray *
 	long row;
 	VteRowData *rowdata;
 	VteScreen *screen;
-	struct vte_charcell cell;
+	VteCell cell;
 
 	screen = terminal->pvt->screen;
 
@@ -3072,20 +2964,19 @@ vte_sequence_handler_screen_alignment_test (VteTerminal *terminal, GValueArray *
 	     row++) {
 		/* Find this row. */
 		while (_vte_ring_next(screen->row_data) <= row)
-			_vte_ring_append(screen->row_data);
+			_vte_terminal_ring_append (terminal, FALSE);
 		_vte_terminal_adjust_adjustments(terminal);
-		rowdata = _vte_ring_index(screen->row_data, row);
+		rowdata = _vte_ring_index_writable (screen->row_data, row);
 		g_assert(rowdata != NULL);
 		/* Clear this row. */
-		if (rowdata->cells->len > 0) {
-			g_array_set_size(rowdata->cells, 0);
-		}
+		_vte_row_data_shrink (rowdata, 0);
+
 		_vte_terminal_emit_text_deleted(terminal);
 		/* Fill this row. */
 		cell.c = 'E';
-		cell.attr = screen->basic_defaults.attr;
+		cell.attr = basic_cell.cell.attr;
 		cell.attr.columns = 1;
-		vte_g_array_fill(rowdata->cells, &cell, terminal->column_count);
+		_vte_row_data_fill (rowdata, &cell, terminal->column_count);
 		_vte_terminal_emit_text_inserted(terminal);
 	}
 	_vte_invalidate_all(terminal);
